'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var _typeof = typeof Symbol === "function" && typeof Symbol.iterator === "symbol" ? function (obj) {
  return typeof obj;
} : function (obj) {
  return obj && typeof Symbol === "function" && obj.constructor === Symbol && obj !== Symbol.prototype ? "symbol" : typeof obj;
};

var PROXY_STATE = Symbol("immer-proxy-state"); // TODO: create per closure, to avoid sharing proxies between multiple immer version
var autoFreeze = true;

function isProxy(value) {
    return !!value && !!value[PROXY_STATE];
}

function isProxyable(value) {
    if (!value) return false;
    if ((typeof value === "undefined" ? "undefined" : _typeof(value)) !== "object") return false;
    if (Array.isArray(value)) return true;
    var proto = Object.getPrototypeOf(value);
    return proto === null || proto === Object.prototype;
}

function freeze(value) {
    if (autoFreeze) {
        Object.freeze(value);
    }
    return value;
}

/**
 * Automatically freezes any state trees generated by immer.
 * This protects against accidental modifications of the state tree outside of an immer function.
 * This comes with a performance impact, so it is recommended to disable this option in production.
 * It is by default enabled.
 *
 * @returns {void}
 */
function setAutoFreeze(enableAutoFreeze) {
    autoFreeze = enableAutoFreeze;
}

// @ts-check

var revocableProxies = null;

var objectTraps = {
    get: get$1,
    has: function has(target, prop) {
        return prop in source(target);
    },
    ownKeys: function ownKeys(target) {
        return Reflect.ownKeys(source(target));
    },

    set: set$1,
    deleteProperty: deleteProperty,
    getOwnPropertyDescriptor: getOwnPropertyDescriptor,
    defineProperty: defineProperty$1,
    setPrototypeOf: function setPrototypeOf() {
        throw new Error("Don't even try this...");
    }
};

var arrayTraps = {};

var _loop = function _loop(key) {
    arrayTraps[key] = function () {
        arguments[0] = arguments[0][0];
        return objectTraps[key].apply(this, arguments);
    };
};

for (var key in objectTraps) {
    _loop(key);
}

function createState(parent, base) {
    return {
        modified: false,
        finalized: false,
        parent: parent,
        base: base,
        copy: undefined,
        proxies: {}
    };
}

function source(state) {
    return state.modified === true ? state.copy : state.base;
}

function get$1(state, prop) {
    if (prop === PROXY_STATE) return state;
    if (state.modified) {
        var value = state.copy[prop];
        if (!isProxy(value) && isProxyable(value)) return state.copy[prop] = createProxy(state, value);
        return value;
    } else {
        if (prop in state.proxies) return state.proxies[prop];
        var _value = state.base[prop];
        if (!isProxy(_value) && isProxyable(_value)) return state.proxies[prop] = createProxy(state, _value);
        return _value;
    }
}

function set$1(state, prop, value) {
    if (!state.modified) {
        if (prop in state.base && Object.is(state.base[prop], value) || prop in state.proxies && state.proxies[prop] === value) return true;
        markChanged(state);
    }
    state.copy[prop] = value;
    return true;
}

function deleteProperty(state, prop) {
    markChanged(state);
    delete state.copy[prop];
    return true;
}

function getOwnPropertyDescriptor(state, prop) {
    var owner = state.modified ? state.copy : prop in state.proxies ? state.proxies : state.base;
    var descriptor = Reflect.getOwnPropertyDescriptor(owner, prop);
    if (descriptor && !(Array.isArray(owner) && prop === "length")) descriptor.configurable = true;
    return descriptor;
}

function defineProperty$1() {
    throw new Error("Immer does currently not support defining properties on draft objects");
}

function markChanged(state) {
    if (!state.modified) {
        state.modified = true;
        state.copy = Array.isArray(state.base) ? state.base.slice() : Object.assign({}, state.base); // TODO: eliminate those isArray checks?
        Object.assign(state.copy, state.proxies); // yup that works for arrays as well
        if (state.parent) markChanged(state.parent);
    }
}

// creates a proxy for plain objects / arrays
function createProxy(parentState, base) {
    var state = createState(parentState, base);
    var proxy = void 0;
    if (Array.isArray(base)) {
        proxy = Proxy.revocable([state], arrayTraps);
    } else {
        proxy = Proxy.revocable(state, objectTraps);
    }
    revocableProxies.push(proxy);
    return proxy.proxy;
}

// given a base object, returns it if unmodified, or return the changed cloned if modified
function finalize(base) {
    if (isProxy(base)) {
        var state = base[PROXY_STATE];
        if (state.modified === true) {
            if (state.finalized === true) return state.copy;
            state.finalized = true;
            if (Array.isArray(state.base)) return finalizeArray(state);
            return finalizeObject(state);
        } else return state.base;
    } else if (base !== null && (typeof base === "undefined" ? "undefined" : _typeof(base)) === "object") {
        // If finalize is called on an object that was not a proxy, it means that it is an object that was not there in the original
        // tree and it could contain proxies at arbitrarily places. Let's find and finalize them as well
        // TODO: optimize this; walk the tree without writing to find proxies
        if (Array.isArray(base)) {
            for (var i = 0; i < base.length; i++) {
                base[i] = finalize(base[i]);
            }return freeze(base);
        }
        var proto = Object.getPrototypeOf(base);
        if (proto === null || proto === Object.prototype) {
            for (var key in base) {
                base[key] = finalize(base[key]);
            }return freeze(base);
        }
    }
    return base;
}

function finalizeObject(state) {
    var copy = state.copy;
    var base = state.base;
    for (var prop in copy) {
        if (copy[prop] !== base[prop]) copy[prop] = finalize(copy[prop]);
    }
    return freeze(copy);
}

function finalizeArray(state) {
    var copy = state.copy;
    var base = state.base;
    for (var i = 0; i < copy.length; i++) {
        if (copy[i] !== base[i]) copy[i] = finalize(copy[i]);
    }
    return freeze(copy);
}

function produceProxy(baseState, producer) {
    var previousProxies = revocableProxies;
    revocableProxies = [];
    try {
        // create proxy for root
        var rootClone = createProxy(undefined, baseState);
        // execute the thunk
        var maybeVoidReturn = producer(rootClone);
        //values either than undefined will trigger warning;
        !Object.is(maybeVoidReturn, undefined) && console.warn("Immer callback expects no return value. However " + (typeof maybeVoidReturn === "undefined" ? "undefined" : _typeof(maybeVoidReturn)) + " was returned");
        // and finalize the modified proxy
        var res = finalize(rootClone);
        // revoke all proxies
        revocableProxies.forEach(function (p) {
            return p.revoke();
        });
        return res;
    } finally {
        revocableProxies = previousProxies;
    }
}

// @ts-check

var descriptors = {};
var states = null;

function createState$1(parent, proxy, base) {
    return {
        modified: false,
        hasCopy: false,
        parent: parent,
        base: base,
        proxy: proxy,
        copy: undefined,
        finished: false,
        finalizing: false
    };
}

function source$1(state) {
    return state.hasCopy ? state.copy : state.base;
}

function _get(state, prop) {
    assertUnfinished(state);
    var value = source$1(state)[prop];
    if (!state.finalizing && !isProxy(value) && isProxyable(value)) {
        prepareCopy(state);
        return state.copy[prop] = createProxy$1(state, value);
    }
    return value;
}

function _set(state, prop, value) {
    assertUnfinished(state);
    if (!state.modified) {
        if (Object.is(source$1(state)[prop], value)) return;
        markChanged$1(state);
    }
    prepareCopy(state);
    state.copy[prop] = value;
}

function markChanged$1(state) {
    if (!state.modified) {
        state.modified = true;
        if (state.parent) markChanged$1(state.parent);
    }
}

function prepareCopy(state) {
    if (state.hasCopy) return;
    state.hasCopy = true;
    state.copy = Array.isArray(state.base) ? state.base.slice() : Object.assign({}, state.base);
}

// creates a proxy for plain objects / arrays
function createProxy$1(parent, base) {
    var proxy = void 0;
    if (Array.isArray(base)) proxy = createArrayProxy(base);else proxy = createObjectProxy(base);
    var state = createState$1(parent, proxy, base);
    createHiddenProperty(proxy, PROXY_STATE, state);
    states.push(state);
    return proxy;
}

function createPropertyProxy(prop) {
    return descriptors[prop] || (descriptors[prop] = {
        configurable: true,
        enumerable: true,
        get: function get$$1() {
            return _get(this[PROXY_STATE], prop);
        },
        set: function set$$1(value) {
            _set(this[PROXY_STATE], prop, value);
        }
    });
}

function createObjectProxy(base) {
    var proxy = Object.assign({}, base);
    Object.keys(base).forEach(function (prop) {
        return Object.defineProperty(proxy, prop, createPropertyProxy(prop));
    });
    return proxy;
}

function createArrayProxy(base) {
    var proxy = new Array(base.length);
    for (var i = 0; i < base.length; i++) {
        Object.defineProperty(proxy, "" + i, createPropertyProxy("" + i));
    }return proxy;
}

function assertUnfinished(state) {
    if (state.finished === true) throw new Error("Cannot use a proxy that has been revoked. Did you pass an object from inside an immer function to an async process?");
}

// this sounds very expensive, but actually it is not that extensive in practice
// as it will only visit proxies, and only do key-based change detection for objects for
// which it is not already know that they are changed (that is, only object for which no known key was changed)
function markChanges() {
    // intentionally we process the proxies in reverse order;
    // ideally we start by processing leafs in the tree, because if a child has changed, we don't have to check the parent anymore
    // reverse order of proxy creation approximates this
    for (var i = states.length - 1; i >= 0; i--) {
        var state = states[i];
        if (state.modified === false) {
            if (Array.isArray(state.base)) {
                if (hasArrayChanges(state)) markChanged$1(state);
            } else if (hasObjectChanges(state)) markChanged$1(state);
        }
    }
}

function hasObjectChanges(state) {
    var baseKeys = Object.keys(state.base);
    var keys = Object.keys(state.proxy);
    return !shallowEqual(baseKeys, keys);
}

function hasArrayChanges(state) {
    return state.proxy.length !== state.base.length;
}

function finalize$1(proxy) {
    // TODO: almost litterally same as Proxy impl; let's reduce code duplication and rollup
    if (isProxy(proxy)) {
        var state = proxy[PROXY_STATE];
        if (state.modified === true) {
            if (Array.isArray(state.base)) return finalizeArray$1(proxy, state);
            return finalizeObject$1(proxy, state);
        } else return state.base;
    } else if (proxy !== null && (typeof proxy === "undefined" ? "undefined" : _typeof(proxy)) === "object") {
        // If finalize is called on an object that was not a proxy, it means that it is an object that was not there in the original
        // tree and it could contain proxies at arbitrarily places. Let's find and finalize them as well
        if (Array.isArray(proxy)) {
            for (var i = 0; i < proxy.length; i++) {
                proxy[i] = finalize$1(proxy[i]);
            }return freeze(proxy);
        }
        var proto = Object.getPrototypeOf(proxy);
        if (proto === null || proto === Object.prototype) {
            for (var key in proxy) {
                proxy[key] = finalize$1(proxy[key]);
            }return freeze(proxy);
        }
    }
    return proxy;
}

function finalizeObject$1(proxy, state) {
    var res = Object.assign({}, proxy);
    var base = state.base;
    for (var prop in res) {
        if (proxy[prop] !== base[prop]) res[prop] = finalize$1(res[prop]);
    }
    return freeze(res);
}

function finalizeArray$1(proxy, state) {
    var res = proxy.slice();
    var base = state.base;
    for (var i = 0; i < res.length; i++) {
        if (res[i] !== base[i]) res[i] = finalize$1(res[i]);
    }
    return freeze(res);
}

function produceEs5(baseState, producer) {
    var prevStates = states;
    states = [];
    try {
        // create proxy for root
        var rootClone = createProxy$1(undefined, baseState);
        // execute the thunk
        var maybeVoidReturn = producer(rootClone);
        //values either than undefined will trigger warning;
        !Object.is(maybeVoidReturn, undefined) && console.warn("Immer callback expects no return value. However " + (typeof maybeVoidReturn === "undefined" ? "undefined" : _typeof(maybeVoidReturn)) + " was returned");
        // and finalize the modified proxy
        for (var i = 0; i < states.length; i++) {
            states[i].finalizing = true;
        } // find and mark all changes (for parts not done yet)
        markChanges();
        var res = finalize$1(rootClone);
        // make sure all proxies become unusable
        for (var _i = 0; _i < states.length; _i++) {
            states[_i].finished = true;
        }return res;
    } finally {
        states = prevStates;
    }
}

function shallowEqual(objA, objB) {
    //From: https://github.com/facebook/fbjs/blob/c69904a511b900266935168223063dd8772dfc40/packages/fbjs/src/core/shallowEqual.js
    if (Object.is(objA, objB)) return true;
    if ((typeof objA === "undefined" ? "undefined" : _typeof(objA)) !== "object" || objA === null || (typeof objB === "undefined" ? "undefined" : _typeof(objB)) !== "object" || objB === null) {
        return false;
    }
    var keysA = Object.keys(objA);
    var keysB = Object.keys(objB);
    if (keysA.length !== keysB.length) return false;
    for (var i = 0; i < keysA.length; i++) {
        if (!hasOwnProperty.call(objB, keysA[i]) || !Object.is(objA[keysA[i]], objB[keysA[i]])) {
            return false;
        }
    }
    return true;
}

function createHiddenProperty(target, prop, value) {
    Object.defineProperty(target, prop, {
        value: value,
        enumerable: false,
        writable: true
    });
}

var useProxies = typeof Proxy !== "undefined";

/**
 * produce takes a state, and runs a function against it.
 * That function can freely mutate the state, as it will create copies-on-write.
 * This means that the original state will stay unchanged, and once the function finishes, the modified state is returned
 *
 * @export
 * @param {any} baseState - the state to start with
 * @param {Function} producer - function that receives a proxy of the base state as first argument and which can be freely modified
 * @returns {any} a new state, or the base state if nothing was modified
 */
function produce(baseState, producer) {
    // curried invocation
    if (arguments.length === 1) {
        var _producer = baseState;
        // prettier-ignore
        if (typeof _producer !== "function") throw new Error("if produce is called with 1 argument, the first argument should be a function");
        return function () {
            var args = arguments;
            return produce(args[0], function (draft) {
                args[0] = draft; // blegh!
                baseState.apply(null, args);
            });
        };
    }

    // prettier-ignore
    {
        if (arguments.length !== 2) throw new Error("produce expects 1 or 2 arguments, got " + arguments.length);
        if (!isProxyable(baseState)) throw new Error("the first argument to produce should be a plain object or array, got " + (typeof baseState === "undefined" ? "undefined" : _typeof(baseState)));
        if (typeof producer !== "function") throw new Error("the second argument to produce should be a function");
    }

    return useProxies ? produceProxy(baseState, producer) : produceEs5(baseState, producer);
}

function setUseProxies(value) {
    useProxies = value;
}

exports['default'] = produce;
exports.setUseProxies = setUseProxies;
exports.setAutoFreeze = setAutoFreeze;
//# sourceMappingURL=immer.js.map
